@@ -4,10 +4,10 @@ from .backbone_autograd import backbone_autograd
 try:
     from .backbone_torch import backbone_torch
 except ImportError:
-    print('Error importing backbone_torch. Possibly pytorch is not installed.')
+    print('Warning: cannot import backbone_torch. Possibly pytorch is not installed.')
 
 
 try:
     from .backbone_jax import backbone_jax
 except ImportError:
-    print('Error importing backbone_jax Possibly JAX is not installed.')
\ No newline at end of file
+    print('Warning: cannot import backbone_jax. Possibly JAX is not installed.')
\ No newline at end of file
