@@ -20,14 +20,14 @@ class Problem:
                 elif backbone == 'jax':
                     # from core.autodiff_jax import autodiff
                     # raise NotImplementedError
-                    from core.backbone_jax import backbone_jax
+                    from ..core.backbone_jax import backbone_jax
                     self.backbone = backbone_jax()
                 elif backbone == 'autograd':
                     # from core.autodiff_ag import autodiff
-                    from core.backbone_autograd import backbone_autograd
+                    from ..core.backbone_autograd import backbone_autograd
                     self.backbone = backbone_autograd()
                 elif backbone == 'torch':
-                    from core.backbone_torch import backbone_torch
+                    from ..core.backbone_torch import backbone_torch
                     self.backbone = backbone_torch(**kwargs)
                 else: 
                     self.backbone = backbone(**kwargs)
